library OpenSpectrumLazarus2014;

{$IF FPC}
{$MODE Delphi}
{$IFEND}

//OpenSpectrumLazarus2014

uses
  SysUtils,
  Windows,
  MATH;
  
const
	kVI_WAVEFORM = $0001; // set if you need the waveform
	kVI_SPECTRUM = $0002; // set if you need the FFT values
	kSONIQUEVISPROC = $0004; // set if you want to allow Soniques user pref vis to affect your vis
						//   for example - blur, smoke and zoom
						
	ALPHA_MASK = $FF000000;
	RED_MASK = $00FF0000;
	GREEN_MASK = $0000FF00;
	BLUE_MASK = $000000FF;

//    207 values
MATRIX: array[0..206] of byte = (
56,190,131,120,114,97,104,98,100,99,94,92,85,98,99,104,105,74,71,85,
77,98,109,87,84,77,115,66,101,106,89,81,79,81,103,102,63,74,78,90,112,
72,82,82,52,55,60,49,48,48,74,37,32,36,37,35,29,32,33,33,31,37,41,40,28,
33,32,36,32,43,26,26,28,28,33,45,31,22,29,30,36,27,27,26,29,33,24,27,36,33,
23,26,27,24,32,28,46,40,30,28,30,44,32,38,34,39,52,44,30,36,47,28,32,37,19,25,
28,26,22,22,21,20,22,18,19,17,17,17,18,17,16,18,17,16,16,17,16,27,15,14,17,17,14,
15,10,13,13,15,15,11,13,14,13,12,10,14,13,12,13,16,13,10,14,12,17,15,15,
16,12,11,11,16,17,18,19,18,15,15,14,10,16,18,10,10,10,14,13,10,18,11,11,
10,11,13,11,11,9,9,11,11,10,9,12,11,10,11,12);

GRAVITY1 = 0.012; // <<  
GRAVITY2 = 0.0005; // <<  

ini_FILE_NAME = 'OpenSpectrumLazarus2014.ini';

type

// >>

TWaveformArray = array [0..1, 0..511] of byte; 
TSpectrumArray = array [0..1, 0..255] of byte;

TScreenArray = array of LongWord;

TVisData = packed record
	MillSec : Longword;
	Waveform: TWaveformArray;
	Spectrum: TSpectrumArray;
end;

TVisColor = LongWord;

pTVisData = ^TVisData;
pInteger = ^Integer;

pPROCEDURE = procedure; cdecl;
pFUNCTION = function: BOOL; cdecl;
pRender = function (Video: Cardinal; width,height,pitch: Integer; VisData: pTVisData): BOOL; cdecl;
pFileProcedure = function(FileName: PChar): BOOL; cdecl;
pClicked = function(x,y,buttons: Integer): BOOL; cdecl;

pReceiveQueryInterface = function(ObjInterface: Pointer):BOOL; cdecl;

TVisInfo = packed record
	Version: LongWord; // 1 = supports Clicked(x,y,buttons) 
	PluginName: PAnsiChar;
	lRequired: Integer;
	
	Initialize: pPROCEDURE;
	Render: pRender;
	SaveSettings: pFileProcedure;
	OpenSettings: pFileProcedure;
	Deinit: pFUNCTION;
	Clicked: pClicked;
	ReceiveQueryInterface: pReceiveQueryInterface;
end;

pTVisInfo = ^TVisInfo;

TScreenData = record
	pixel: Cardinal;
	Width: Integer;
	Height: Integer;
	Pitch: Integer;
end;
	
// DLL CODE

TSingleArray = array [0..206] of single;

TThemeColor = record 
	BAR: TVisColor;
	PEAK: TVisColor;
	BKG: TVisColor;
end;

var 
	DATA: pTVisInfo;
	DATA_NAME: AnsiString;
    SP_MATRIX: array [0..206] of byte; // <<   
    SP_DATA: TSingleArray; // <<   
    //    
    BAR_SPEED: TSingleArray; //  
    BAR_VALUE: TSingleArray; //  
	PEAK_SPEED: TSingleArray; //  
	PEAK_VALUE: TSingleArray; //  
    // 
	
	THEME: array of TThemeColor;
    S_THEME: byte;
	
// DRAW HELPER Functions
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

procedure DS_PIXEL(const DS: TScreenData; x,y: Integer; Color: TVisColor);
var z: cardinal;
type pLongWord = ^LongWord;
begin
	with DS do begin
        z:=pixel+cardinal((Pitch*y+x)*SizeOf(LongWord));
        pLongWord(z)^:=Color;
        // C++ low level pointer operations :D
    end;
end;

procedure DS_DRAW_BAND(const DS: TScreenData; X,Height,Width: Integer; Color: TVisColor);
var i,j: integer;
begin
	//  
	for i:=0 to Height-1 do
		for j:=0 to Width-1 do 
			DS_PIXEL(DS, X+j, DS.Height-i, Color);
end;

procedure DS_DRAW_LINE(const DS: TScreenData; X,Height,Width: Integer; Color: TVisColor);
var j: integer;
begin
    for j:=0 to Width-1 do
       DS_PIXEL(DS, X+j, DS.Height-Height, Color);
end;

// DLL CODE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
	
procedure Initialize; cdecl;
var i: integer;
f: FILE of byte; theme_num: byte;
begin
   for i:=0 to High(SP_MATRIX) do begin
      SP_MATRIX[i]:=round(MATRIX[i]*0.5);
      BAR_SPEED[i]:=0;
	  BAR_VALUE[i]:=0;
	  PEAK_SPEED[i]:=0;
	  PEAK_VALUE[i]:=0;
   end;

   //   

   SetLength(THEME,3);

   //RGB() -> BLUE GREEN RED     

   with THEME[0] do begin  // AIMP 
     BKG:=RGB($18,$18,$18);
     BAR:=RGB(50,151,255);
     PEAK:=RGB($CC,$CC,$CC);
   end;

   with THEME[1] do begin  //  
     BKG:=0;
     BAR:=GREEN_MASK;
     PEAK:=RGB($CC,$CC,$CC);
   end;

   with THEME[2] do begin //  
     BKG:=0;
     BAR:=RGB($FF,$33,$0A);
     PEAK:=RGB($CC,$CC,$CC);
   end;

   // <<

   if not FileExists(ini_FILE_NAME) then begin
       Assign(f,ini_FILE_NAME);
       Rewrite(f);
       theme_num:=0;
       Write(f,theme_num);
       Close(f);
   end;

   Assign(f,ini_FILE_NAME);
   Reset(f);
   read(f,S_THEME);

   if (S_THEME>High(THEME)) then  S_THEME:=High(THEME);

   Close(f);

end;

procedure ZipSpectrum(var i,z:Integer; limit, blockSize: Integer; hpower: Single);
//   
var 
	h,hmax: Single;
	j: Integer;
begin
	while (i<limit) do begin
       h:=0;
	   hmax:=0;
       for j:=1 to blockSize do begin
           if hmax<SP_DATA[i] then hmax:=SP_DATA[i];
		   h:=h+SP_DATA[i];
           inc(i);
       end;
       h:=h/blockSize*0.2 + hmax*0.8;
       //h:=hmax;
       h:=power(h,hpower);
       SP_DATA[z] := h;
       inc(z);
    end;
end;

procedure GRAVITYSprectrum(var VALUE,SPEED,ORIGINAL: TSingleArray; i: Integer; GRAVITY_SPEED: Single);
//   
begin
	if VALUE[i]<ORIGINAL[i] then begin
		VALUE[i]:=ORIGINAL[i];
		SPEED[i]:=0;
	end else begin
		SPEED[i]:=SPEED[i]+GRAVITY_SPEED;
		VALUE[i]:=VALUE[i]-SPEED[i];
		if (VALUE[i]<0) then begin
			VALUE[i]:=0;
			SPEED[i]:=0;
		end;
	end;
end;

function Render(Video: Cardinal; width,height,pitch: Integer; VisData: pTVisData): BOOL; cdecl;
var 
	color,color_BAR,color_PEAK: TVisColor;
	i,j,z: integer;
	h: single;
    hpower: single;
    DS: TScreenData;
begin
// DS INIT
	DS.pixel := Video;
	DS.Width := width;
	DS.Height := height;
	DS.Pitch := pitch;

// DRAW
    color:=THEME[S_THEME].BKG;
    for i:=0 to height-1 do     //   
       for j:=0 to width-1 do
          DS_PIXEL(DS,j,i,color);

    //   

    for i:=0 to High(SP_MATRIX) do begin
        z:= VisData^.Spectrum[0][i];
        //  1  
		if SP_MATRIX[i]<z then SP_MATRIX[i]:=z;
        //  2  
        SP_DATA[i]:= z / SP_MATRIX[i] ;
    end;

    {
       41 (4   ,  165 )
       - 207 
     0.3 * 207 = 62  (30%)
     62    -   25  (~3*25 = 75)
     0.7 * 207 = 145 (70%)
     145      16   (~8*16 = 128)
    }
    i:=0; z:=0;
	hpower:=0.75;
	
	//   3*25,   3 
	ZipSpectrum(i,z,3*25,3,hpower);
	
	//   207-8,   8 
	ZipSpectrum(i,z,207-8,8,hpower);
	
	//  3  
	
	for i:=0 to z-1 do begin 
		// 
		GRAVITYSprectrum(BAR_VALUE,BAR_SPEED,SP_DATA,i,GRAVITY1);
		// 
		GRAVITYSprectrum(PEAK_VALUE,PEAK_SPEED,BAR_VALUE,i,GRAVITY2);
	end;
	
	//  4 
    color_BAR:=THEME[S_THEME].BAR;
    color_PEAK:=THEME[S_THEME].PEAK;

	i:=0;
    h:=abs(height)*0.85;
    if h>110 then h:=110;
    j:=round(width/z);
	while (i*j<width) AND (i<z) do begin
		// 
		DS_DRAW_BAND(DS,i*j, trunc(BAR_VALUE[i]*h),j-1, color_BAR);
		// 
		DS_DRAW_LINE(DS,i*j, trunc(PEAK_VALUE[i]*h),j-1, color_PEAK);
		
		inc(i);
	end;
end;

function Clicked(x,y,buttons: Integer): BOOL; cdecl;
var f: file of byte;
begin
   //   
   S_THEME:=S_THEME+1;
   if (S_THEME>High(THEME)) then S_THEME:=0;
   Assign(f,ini_FILE_NAME);
   Rewrite(f);
   write(f,S_THEME);
   Close(f);
   result:=true;
end;

// TRASH Functions
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	

function Deinit: BOOL; cdecl;
begin
	Dispose(DATA);
	DATA:=nil;
end;

function SaveSettings(FileName: PChar): BOOL; cdecl;
begin
    result:=true;
end;

function OpenSettings(FileName: PChar): BOOL; cdecl;
begin
    result:=true;
end;

function ReceiveQueryInterface(ObjInterface: Pointer):BOOL; cdecl;
begin
	result:=true;
end;

// DLL CODE Export
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	
function QueryModule: pTVisInfo; cdecl;
begin
	//   DLL
	new(DATA);
	DATA_NAME := 'OpenSpectrumLazarus2014';
	
	with DATA^ do begin
		Version := 1;//        function Clicked(x,y,buttons: Integer): BOOL
		PluginName := PAnsiChar(DATA_NAME);
		lRequired := kVI_SPECTRUM or kSONIQUEVISPROC;
	end;

    DATA^.Initialize:= @Initialize;
    DATA^.Render:= @Render;
    DATA^.SaveSettings:= @SaveSettings;
    DATA^.OpenSettings:= @OpenSettings;
    DATA^.Deinit:= @Deinit;
    DATA^.Clicked:= @Clicked;
    DATA^.ReceiveQueryInterface:= @ReceiveQueryInterface;

    result:=DATA;
end;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DLL READY
exports QueryModule;

begin

end.


